/**
 * @file
 * Ethernet Interface for standalone applications (without RTOS) - works only for 
 * ethernet polling mode (polling for ethernet frame reception)
 *
 */

/*
 * Copyright (c) 2001-2004 Swedish Institute of Computer Science.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification,
 * are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT
 * SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
 * IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
 * OF SUCH DAMAGE.
 *
 * This file is part of the lwIP TCP/IP stack.
 *
 * Author: Adam Dunkels <adam@sics.se>
 *
 */

#include "lwip/mem.h"
#include "netif/etharp.h"
#include "ethernetif.h"
#include "at32f403a_407_emac.h"
#include "common.h"
#include "common_config.h"   
#include "settings_api.h"
#include "utility.h"
#include <string.h>

// Network interface name
#define IFNAME0 's'
#define IFNAME1 't'

#define EMAC_DMARxDesc_FrameLengthShift  16

/**
 * Helper struct to hold private data used to operate your ethernet interface.
 * Keeping the ethernet address of the MAC in this struct is not necessary
 * as it is already kept in the struct netif.
 * But this is only an example, anyway...
 */
struct ethernetif
{
  struct eth_addr *ethaddr;
  /* Add whatever per-interface state that is needed here. */
  int unused;
};

#define EMAC_RXBUFNB        4
#define EMAC_TXBUFNB        2


// Ethernet Rx & Tx DMA Descriptors
#pragma data_alignment = 4
emac_dma_desc_type  DMARxDscrTab[EMAC_RXBUFNB];
#pragma data_alignment = 4
emac_dma_desc_type  DMATxDscrTab[EMAC_TXBUFNB];
#pragma data_alignment = 4
uint8_t Rx_Buff[EMAC_RXBUFNB][EMAC_MAX_PACKET_LENGTH]; 
#pragma data_alignment = 4
uint8_t Tx_Buff[EMAC_TXBUFNB][EMAC_MAX_PACKET_LENGTH];

extern emac_dma_desc_type  *dma_tx_desc_to_set;
extern emac_dma_desc_type  *dma_rx_desc_to_get;

typedef struct
{
    u32 length;
    u32 buffer;
    emac_dma_desc_type *descriptor;
    
} FrameTypeDef;

FrameTypeDef emac_rxpkt_chainmode(void);
u32 emac_getcurrenttxbuffer(void);
error_status emac_txpkt_chainmode(u16 FrameLength);


#ifdef ETHERNETIF_RTOS

#include "FreeRTOS.h"
#include "task.h"

SemaphoreHandle_t PHY_RX_xSemaphore;
SemaphoreHandle_t PHY_TX_xSemaphore;

#define PHY_PRIO            (configMAX_PRIORITIES - 1)
#define PHY_STK_SIZE        1024

extern struct netif xnetif;
struct netif *lwip_netif= &xnetif;

TaskHandle_t PHY_Handler;

void network_task(void *params);


#endif



/**
 * In this function, the hardware should be initialized.
 * Called from ethernetif_init().
 *
 * @param netif the already initialized lwip network interface structure
 *        for this ethernetif
 */
static void low_level_init(struct netif *netif)
{
    uint8_t mac[6];
    
    // set MAC hardware address length
    netif->hwaddr_len = ETHARP_HWADDR_LEN;
    
    SETTINGS_GetMac(mac);
    
    emac_local_address_set(mac);
    
    netif->hwaddr[0] =  mac[0];
    netif->hwaddr[1] =  mac[1];
    netif->hwaddr[2] =  mac[2];
    netif->hwaddr[3] =  mac[3];
    netif->hwaddr[4] =  mac[4];
    netif->hwaddr[5] =  mac[5];
    
    // maximum transfer unit
    netif->mtu = 1500;
    
    // device capabilities
    // don't set NETIF_FLAG_ETHARP if this device is not an ethernet one
    netif->flags = NETIF_FLAG_BROADCAST | NETIF_FLAG_ETHARP | NETIF_FLAG_LINK_UP;
    
    // Initialize Tx Descriptors list: Chain Mode
    emac_dma_descriptor_list_address_set(EMAC_DMA_TRANSMIT, DMATxDscrTab, &Tx_Buff[0][0], EMAC_TXBUFNB);
    // Initialize Rx Descriptors list: Chain Mode 
    emac_dma_descriptor_list_address_set(EMAC_DMA_RECEIVE, DMARxDscrTab, &Rx_Buff[0][0], EMAC_RXBUFNB);


#ifdef ETHERNETIF_RTOS
  
    static bool task_eth_flag = false;
  
    if (PHY_RX_xSemaphore == NULL) {
        PHY_RX_xSemaphore = xSemaphoreCreateCounting(20,0);
    }
    
/*    
    if (PHY_TX_xSemaphore == NULL) {
        PHY_TX_xSemaphore = xSemaphoreCreateBinary();
    }
*/
    
    if (!task_eth_flag) 
    {
        xTaskCreate((TaskFunction_t)network_task, (const char*)"PHY",(uint16_t)PHY_STK_SIZE, 
                    (void*)NULL, (UBaseType_t)PHY_PRIO, (TaskHandle_t*)&PHY_Handler); 
        
        task_eth_flag = true;
    }
  
#endif  
  
  
  
  // Enable MAC and DMA transmission and reception
  emac_start();
}

/**
 * This function should do the actual transmission of the packet. The packet is
 * contained in the pbuf that is passed to the function. This pbuf
 * might be chained.
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @param p the MAC packet to send (e.g. IP packet including MAC addresses and type)
 * @return ERR_OK if the packet could be sent
 *         an err_t value if the packet couldn't be sent
 *
 * @note Returning ERR_MEM here if a DMA queue of your MAC is full can lead to
 *       strange results. You might consider waiting for space in the DMA queue
 *       to become availale since the stack doesn't retry to send a packet
 *       dropped because of memory failure (except for the TCP timers).
 */

static err_t low_level_output(struct netif *netif, struct pbuf *p)
{
  (void)netif;
  struct pbuf *q;
  int framelength = 0;
  u8 *buffer =  (u8 *)(dma_tx_desc_to_set->buf1addr);
  
  /* copy frame from pbufs to driver buffers */
  for(q = p; q != NULL; q = q->next) 
  {
    memcpy((u8_t*)&buffer[framelength], q->payload, q->len);
	framelength = framelength + q->len;
  }
  
#if defined (TRAFFIC_ENABLE)  
  trf_inc(TR_ETH_TX, framelength); // трафик (tx)
#endif  
  /* Note: padding and CRC for transmitted frame 
     are automatically inserted by DMA */
    
  /* Prepare transmit descriptors to give to DMA*/ 
  if(emac_txpkt_chainmode(framelength) == ERROR)
  {
    return ERR_MEM;
  }

  return ERR_OK;
}

/**
 * Should allocate a pbuf and transfer the bytes of the incoming
 * packet from the interface into the pbuf.
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @return a pbuf filled with the received packet (including MAC header)
 *         NULL on memory error
 */
static struct pbuf * low_level_input(struct netif *netif)
{
  (void)netif;
  struct pbuf *p, *q;
  u16_t len;
  int l =0;
  FrameTypeDef frame;
  u8 *buffer;
  
  p = NULL;
  
  /* get received frame */
  frame = emac_rxpkt_chainmode();
  
  /* Obtain the size of the packet and put it into the "len" variable. */
  len = frame.length;
  buffer = (u8 *)frame.buffer;
  
  /* We allocate a pbuf chain of pbufs from the Lwip buffer pool */
  p = pbuf_alloc(PBUF_RAW, len, PBUF_POOL);
  
  /* copy received frame to pbuf chain */
  if (p != NULL)
  {
    for (q = p; q != NULL; q = q->next)
    {
      memcpy((u8_t*)q->payload, (u8_t*)&buffer[l], q->len);
      l = l + q->len;
    }
#if defined (TRAFFIC_ENABLE)    
    trf_inc(TR_ETH_RX, l); // трафик (rx)
#endif    
  }
  
  /* Set Own bit of the Rx descriptor Status: gives the buffer back to ETHERNET DMA */
  frame.descriptor->status = EMAC_DMARXDESC_OWN;

  /* When Rx Buffer unavailable flag is set: clear it and resume reception */
  if(emac_dma_flag_get(EMAC_DMA_RBU_FLAG))
  {
    /* Clear RBUS ETHERNET DMA flag */
    emac_dma_flag_clear(EMAC_DMA_RBU_FLAG);
    /* Resume DMA reception */
    EMAC_DMA->rpd_bit.rpd = FALSE;
  }
  return p;
}

/**
 * This function should be called when a packet is ready to be read
 * from the interface. It uses the function low_level_input() that
 * should handle the actual reception of bytes from the network
 * interface. Then the type of the received packet is determined and
 * the appropriate input function is called.
 *
 * @param netif the lwip network interface structure for this ethernetif
 */
err_t ethernetif_input(struct netif *netif)
{
  err_t err;
  struct pbuf *p;

  /* move received packet into a new pbuf */
  p = low_level_input(netif);

  /* no packet could be read, silently ignore this */
  if (p == NULL) return ERR_MEM;

  /* entry point to the LwIP stack */
  err = netif->input(p, netif);
  
  if (err != ERR_OK)
  {
    LWIP_DEBUGF(NETIF_DEBUG, ("ethernetif_input: IP input error\n"));
    pbuf_free(p);
    p = NULL;
  }
  return err;
}

/**
 * Should be called at the beginning of the program to set up the
 * network interface. It calls the function low_level_init() to do the
 * actual setup of the hardware.
 *
 * This function should be passed as a parameter to netif_add().
 *
 * @param netif the lwip network interface structure for this ethernetif
 * @return ERR_OK if the loopif is initialized
 *         ERR_MEM if private data couldn't be allocated
 *         any other err_t on error
 */
err_t ethernetif_init(struct netif *netif)
{
  LWIP_ASSERT("netif != NULL", (netif != NULL));
  
#if LWIP_NETIF_HOSTNAME
  /* Initialize interface hostname */
  netif->hostname = "lwip";
#endif /* LWIP_NETIF_HOSTNAME */

  netif->name[0] = IFNAME0;
  netif->name[1] = IFNAME1;
  /* We directly use etharp_output() here to save a function call.
   * You can instead declare your own function an call etharp_output()
   * from it if you have to do some checks before sending (e.g. if link
   * is available...) */
  netif->output = etharp_output;
  netif->linkoutput = low_level_output;

  /* initialize the hardware */
  low_level_init(netif);

  return ERR_OK;
}


/*******************************************************************************
* Function Name  : emac_rxpkt_chainmode
* Description    : Receives a packet.
* Input          : None
* Output         : None
* Return         : frame: farme size and location
*******************************************************************************/
FrameTypeDef emac_rxpkt_chainmode(void)
{
  u32 framelength = 0;
  FrameTypeDef frame = {0,0};

  /* Check if the descriptor is owned by the ETHERNET DMA (when set) or CPU (when reset) */
  if((dma_rx_desc_to_get->status & EMAC_DMARXDESC_OWN) != (u32)RESET)
  {
    frame.length = FALSE;

    if(emac_dma_flag_get(EMAC_DMA_RBU_FLAG))
    {
      /* Clear RBUS ETHERNET DMA flag */
      emac_dma_flag_clear(EMAC_DMA_RBU_FLAG);
      /* Resume DMA reception */
      EMAC_DMA->rpd_bit.rpd = FALSE;
    }
    /* Return error: OWN bit set */
    return frame;
  }

  if(((dma_rx_desc_to_get->status & EMAC_DMATXDESC_ES) == (u32)RESET) &&
     ((dma_rx_desc_to_get->status & EMAC_DMARXDESC_LS) != (u32)RESET) &&
     ((dma_rx_desc_to_get->status & EMAC_DMARXDESC_FS) != (u32)RESET))
  {
    /* Get the Frame Length of the received packet: substruct 4 bytes of the CRC */
    framelength = ((dma_rx_desc_to_get->status & EMAC_DMARXDESC_FL) >> EMAC_DMARxDesc_FrameLengthShift) - 4;

    /* Get the addrees of the actual buffer */
    frame.buffer = dma_rx_desc_to_get->buf1addr;
  }
  else
  {
    /* Return ERROR */
    framelength = FALSE;
  }

  frame.length = framelength;

  frame.descriptor = dma_rx_desc_to_get;

  /* Update the ETHERNET DMA global Rx descriptor with next Rx decriptor */
  /* Chained Mode */
  /* Selects the next DMA Rx descriptor list for next buffer to read */
  dma_rx_desc_to_get = (emac_dma_desc_type*) (dma_rx_desc_to_get->buf2nextdescaddr);
  /* Return Frame */
  return (frame);
}

/*******************************************************************************
* Function Name  : emac_txpkt_chainmode
* Description    : Transmits a packet, from application buffer, pointed by ppkt.
* Input          : - FrameLength: Tx Packet size.
* Output         : None
* Return         : ERROR: in case of Tx desc owned by DMA
*                  SUCCESS: for correct transmission
*******************************************************************************/
error_status emac_txpkt_chainmode(u16 FrameLength)
{
  /* Check if the descriptor is owned by the ETHERNET DMA (when set) or CPU (when reset) */
  if((dma_tx_desc_to_set->status & EMAC_DMATXDESC_OWN) != (u32)RESET)
  {
    /* Return ERROR: OWN bit set */
    return ERROR;
  }

  /* Setting the Frame Length: bits[12:0] */
  dma_tx_desc_to_set->controlsize = (FrameLength & EMAC_DMATXDESC_TBS1);

  /* Setting the last segment and first segment bits (in this case a frame is transmitted in one descriptor) */
  dma_tx_desc_to_set->status |= EMAC_DMATXDESC_LS | EMAC_DMATXDESC_FS;

  /* Set Own bit of the Tx descriptor Status: gives the buffer back to ETHERNET DMA */
  dma_tx_desc_to_set->status |= EMAC_DMATXDESC_OWN;
  /* When Tx Buffer unavailable flag is set: clear it and resume transmission */
  if(emac_dma_flag_get(EMAC_DMA_TBU_FLAG))
  {
    /* Clear TBUS ETHERNET DMA flag */
    emac_dma_flag_clear(EMAC_DMA_TBU_FLAG);
    /* Resume DMA transmission*/
    EMAC_DMA->tpd_bit.tpd = 0;
  }

  /* Update the ETHERNET DMA global Tx descriptor with next Tx decriptor */
  /* Chained Mode */
  /* Selects the next DMA Tx descriptor list for next buffer to send */
  dma_tx_desc_to_set = (emac_dma_desc_type*) (dma_tx_desc_to_set->buf2nextdescaddr);
  /* Return SUCCESS */
  
  return SUCCESS;
}


#ifdef ETHERNETIF_RTOS

//
void network_task(void *params)
{
    struct pbuf *p;
    err_t err;
    
    for (;;)
    {
        if (xSemaphoreTake(PHY_RX_xSemaphore, portMAX_DELAY) == pdTRUE)
        {
            p = low_level_input(lwip_netif);
        }
        
        if (p != NULL)
        {
            err = lwip_netif->input(p, lwip_netif);
        
            if (err != ERR_OK)
            {
                pbuf_free(p);
                p = NULL;
            }
        
        }
    }
}
#endif

